"""Status objects for veliberator"""
from datetime import datetime, timedelta
from xml.dom.minidom import parseString
from xml.parsers.expat import ExpatError

from veliberator.settings import STATION_STATUS_RECENT
from veliberator.settings import XML_URL_STATUS_STATION
from veliberator.xml_wrappers import xml_station_status_wrapper
from veliberator.grabber import Grabber

global_stationstatus_cache = {}

def cache_wrapper(method):
    def cache(instance):
        """Use a timed cache for result, and set
        the results into the instance"""
        global global_stationstatus_cache
        
        key_cache = instance.velib_id
        if not global_stationstatus_cache.has_key(key_cache) or \
               global_stationstatus_cache[key_cache]['datetime'] + \
               timedelta(minutes=STATION_STATUS_RECENT) < datetime.now():
            global_stationstatus_cache[key_cache] = method(instance)
        instance.status = global_stationstatus_cache[key_cache]
        return instance.status
    return cache

class StationStatus(object):
    """Status of a station, by opening an url
    providing a xml file, with cache"""
    status = {}

    def __init__(self, velib_id, xml_url=XML_URL_STATUS_STATION):
        """Init the status"""
        self.velib_id = velib_id
        self.xml_url = xml_url + str(self.velib_id)
        self.get_status()

    @cache_wrapper
    def get_status(self):
        """Get the status provided by an URL"""
        try:
            dom = parseString(Grabber(self.xml_url).content)        
            status = xml_station_status_wrapper(dom.firstChild)
        except (IOError, IndexError, ValueError, ExpatError):
            status = {'total': 0, 'available': 0,
                      'free': 0, 'ticket': False}
        status['datetime'] = datetime.now()
        return status

    def __getattr__(self, name):
        """Allow direct access to self.status items"""
        if self.status.has_key(name):
            return self.status.get(name)
        return getattr(self, name)

    def __repr__(self):
        return '<StationStatus "%s">' % self.velib_id



#!/usr/bin/python
import sys, os
import unittest

sys.path.append(os.path.abspath('.'))

from tests import global_test_suite

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(global_test_suite)
    



"""Global suite of tests"""
import unittest

from veliberator.models import db_connection

from tests.test_models import suite as models_suite
from tests.test_xml_wrappers import suite as xml_suite
from tests.test_station import suite as station_suite
from tests.test_cartography import suite as cartography_suite
from tests.test_status import suite as status_suite
from tests.test_geofinder import suite as geofinder_suite
from tests.test_grabber import suite as grabber_suite

db_connection('sqlite:///:memory:')

global_test_suite = unittest.TestSuite([models_suite,
                                        xml_suite,
                                        station_suite,
                                        cartography_suite,
                                        status_suite,
                                        geofinder_suite,
                                        grabber_suite,])





#!/usr/bin/python
import sys, os
from optparse import OptionParser

sys.path.append(os.path.abspath('.'))

from veliberator import Cartography
from veliberator.models import db_connection
from veliberator.settings import DATABASE_URI

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option('-d', '--database', dest='database',
                      help='The SQLURI of the database', default=DATABASE_URI,
                      metavar='DATABASE')
    (options, args) = parser.parse_args()
    
    db_connection(options.database)
    print 'Database open, now synchronizing...'
    Cartography.synchronize()
    print 'Synchronization completed !'
    

    
    
    



"""Settings for Veliberator"""
import os
from ConfigParser import SafeConfigParser
from ConfigParser import NoSectionError, NoOptionError

config = SafeConfigParser()
config.read([os.path.join('/etc', 'veliberator.cfg'),
             os.path.expanduser('~/.veliberator.cfg'),
             os.path.join('etc', 'veliberator.cfg'),])

XML_URL_DATA_STATION = config.get('XML', 'url_data_station')
XML_URL_STATUS_STATION = config.get('XML', 'url_status_station')

STATION_STATUS_RECENT = config.getint('STATION', 'status_recent')
STATION_ALMOST_FULL = config.getint('STATION', 'almost_full')
STATION_ALMOST_EMPTY = config.getint('STATION', 'almost_empty')

DATABASE_URI = config.get('DATABASE', 'uri')
DATABASE_ECHO = config.getboolean('DATABASE', 'echo')

TEST_XML_URL_DATA_STATION = config.get('TEST', 'url_data_station')

PROXY_SERVERS = config.get('PROXY', 'servers')
PROXY_SERVERS = PROXY_SERVERS and PROXY_SERVERS.split(';') or []



"""Unit tests for Cartography object"""
import unittest

from veliberator.settings import TEST_XML_URL_DATA_STATION
from veliberator.cartography import Cartography
from veliberator.models import StationInformation

class CartographyTestCase(unittest.TestCase):

    def test_SynchronizeFlush(self):
        self.assertEquals(StationInformation.query.count(), 0)

        Cartography.synchronize(TEST_XML_URL_DATA_STATION)
        self.assertEquals(StationInformation.query.count(), 112)

        Cartography.flush()
        self.assertEquals(StationInformation.query.count(), 0)

    
suite = unittest.TestLoader().loadTestsFromTestCase(CartographyTestCase)



"""Station objects for veliberator"""
from xml.dom.minidom import parse

from veliberator.models import db_connection
from veliberator.models import StationInformation
from veliberator.status import StationStatus
from veliberator.geofinder import StationGeoFinder
from veliberator.settings import STATION_ALMOST_FULL
from veliberator.settings import STATION_ALMOST_EMPTY

STATUS_OPEN = u'open'
STATUS_CLOSE = u'close'
STATUS_BONUS = u'bonus'
STATUS_ERROR = u'api-error'
STATUS_BIKE_ONLY = u'bike-only'
STATUS_PARKING_ONLY = u'parking-only'
STATUS_ALMOST_FULL = u'almost-full'
STATUS_ALMOST_EMPTY = u'almost-empty'
STATUS_NO_SERVICE = u'no-service'

class UnknowStation(Exception):
    pass

class Station(object):
    """Station object,
    combinating informations and status"""

    def __init__(self, velib_id):
        self.id = int(velib_id)
        self.informations = self.acquire_informations()

        if not self.informations:
            raise UnknowStation('The Station ID does not exist.')

        self.status = StationStatus(self.id)
        self.finder = StationGeoFinder(self)

    def acquire_informations(self):
        """Default method for acquiring informations
        of a stations, by connecting to the database"""
        try:
            return StationInformation.get(self.id)
        except AttributeError:
            db_connection()
            return StationInformation.get(self.id)

    @property
    def is_open(self):
        """Property telling if the station is open"""
        return self.informations.opened

    @property
    def is_bonus(self):
        """Property telling if the station is bonus"""
        return self.informations.bonus

    def is_free(self, places=1):
        """Method property if the station
        has free places"""
        return self.is_open and self.status.free >= places

    def is_available(self, places=1):
        """Method property if the station
        has available places"""
        return self.is_open and self.status.available >= places

    @property
    def stations_around(self):
        """Find the stations around"""
        return self.finder.get_stations_around()

    @property
    def state(self):
        """Return the status for the station"""
        if not self.is_open:
            return STATUS_CLOSE

        free = self.status.free
        available = self.status.available
        total =  self.status.total

        if not available and not free:
            if not total:
                return STATUS_ERROR
            return STATUS_NO_SERVICE
        elif available and not free:
            return STATUS_BIKE_ONLY
        elif not available and free:
            return STATUS_PARKING_ONLY
        elif STATION_ALMOST_FULL >= free:
            return STATUS_ALMOST_FULL
        elif STATION_ALMOST_EMPTY >= available:
            return STATUS_ALMOST_EMPTY
        elif self.is_bonus:
            return STATUS_BONUS
        return STATUS_OPEN

    def __repr__(self):
        if self.informations:
            return '<Station "%s" (%s)>' % (self.id, self.informations.address)
        return '<Station "%s">' % self.id





"""Unit tests for GeoFinder objects"""
import unittest

from veliberator.geofinder import global_geofinder_cache
from veliberator.geofinder import BaseGeoFinder
from veliberator.geofinder import StationGeoFinder
from veliberator.geofinder import AddressGeoFinder
from veliberator.geofinder import GeoFinderError
from veliberator.geofinder import pythagor_distance
from veliberator.geofinder import haversine_distance
from veliberator.models import StationInformation
from veliberator import Cartography
from veliberator import Station
from veliberator.settings import TEST_XML_URL_DATA_STATION

class BaseGeoFinderTestCase(unittest.TestCase):

    def test_ComputeSquareArea(self):
        finder = BaseGeoFinder(1, 1)
        self.assertEquals(finder.compute_square_area(),
                          (('0.99', '1.00', '1.01'), ('0.99', '1.00', '1.01')))
        finder.lat = '1'
        finder.lng = '1'
        self.assertEquals(finder.compute_square_area(),
                          (('0.99', '1.00', '1.01'), ('0.99', '1.00', '1.01')))
        finder.lat = 4.897645
        finder.lng = 18.798923
        self.assertEquals(finder.compute_square_area(),
                          (('4.89', '4.90', '4.91'), ('18.79', '18.80', '18.81')))

    def test_ComputeStationDistances(self):
        finder = BaseGeoFinder(1, 1)

        si1 = StationInformation(id=42, lat=1, lng=1)
        result = finder.compute_station_distances([si1])
        self.assertEquals(result, [])

        si2 = StationInformation(id=43, lat=4, lng=5)
        result = finder.compute_station_distances([si2,])
        self.assertEquals(result, [si2])
        self.assertEquals(result[0].distance, 556281.9630214232)
        result = finder.compute_station_distances([si2,], pythagor_distance)
        self.assertEquals(result[0].distance, 5.0)

        si3 = StationInformation(id=44, lat=8, lng=10)
        si4 = StationInformation(id=45, lat=6, lng=5.5)
        result = finder.compute_station_distances([si1, si2, si3, si4])
        self.assertEquals(len(result), 3)

        StationInformation.query.delete()

    def test_GetStationsInArea(self):
        test_1 = (('4.89', '4.90', '4.91'), ('18.79', '18.80', '18.81'))
        test_2 = (('48.80', '48.81', '48.82'), ('2.37', '2.38', '2.39'))

        Cartography.synchronize(TEST_XML_URL_DATA_STATION)
        finder = BaseGeoFinder(1, 1)

        self.assertEquals(len(finder.get_stations_in_area(*test_1)), 0)
        self.assertEquals(len(finder.get_stations_in_area(*test_2)), 8)
                
        Cartography.flush()

    def test_GetStationsAround(self):
        Cartography.synchronize(TEST_XML_URL_DATA_STATION)

        finder = BaseGeoFinder(1, 1)
        self.assertEquals(len(finder.get_stations_around()), 0)

        finder.lat = 48.81
        finder.lng = 2.38
        self.assertEquals([station.id for station in finder.get_stations_around()],
                          [42012, 42010, 42009, 42008, 42016, 42006, 42007, 42015])
        
        Cartography.flush()

    def test_Cache(self):
        key = (1, 1)
        self.assertFalse(global_geofinder_cache.has_key(key))        
        finder = BaseGeoFinder(key[0], key[1])
        finder.get_stations_around()
        self.assertTrue(global_geofinder_cache.has_key(key))


class StationGeoFinderTestCase(unittest.TestCase):

    def test_Init(self):
        Cartography.synchronize(TEST_XML_URL_DATA_STATION)

        station = Station(42008)
        geofinder = StationGeoFinder(station)
        self.assertEquals(geofinder.lat, station.informations.lat)
        self.assertEquals(geofinder.lng, station.informations.lng)

        Cartography.flush()

class AddressGeoFinderTestCase(unittest.TestCase):

    def test_Init(self):
        address = '1 place de la Bastille, 75012 Paris'
        geofinder = AddressGeoFinder(address)
        self.assertEquals(geofinder.precision, 8)
        self.assertEquals(geofinder.lat, 48.853085399999998)
        self.assertEquals(geofinder.lng, 2.3687089000000001)

        address = 'Bethune, 62400'
        self.assertRaises(GeoFinderError, AddressGeoFinder, address)

    def test_GeoCompute(self):
        address = '1 place de la Bastille, 75012 Paris'
        geofinder = AddressGeoFinder(address)
        self.assertEquals(len(geofinder.geocompute(address)), 5)

suite = unittest.TestSuite([
    unittest.TestLoader().loadTestsFromTestCase(BaseGeoFinderTestCase),
    unittest.TestLoader().loadTestsFromTestCase(StationGeoFinderTestCase),
    unittest.TestLoader().loadTestsFromTestCase(AddressGeoFinderTestCase),])




"""Objects for helping filling data"""
from xml.dom.minidom import parseString

from veliberator.grabber import Grabber
from veliberator.settings import XML_URL_DATA_STATION
from veliberator.xml_wrappers import xml_station_information_wrapper
from veliberator.models import StationInformation, session


class Cartography(object):
    """Grab the data and save it in db"""

    @staticmethod
    def synchronize(url=XML_URL_DATA_STATION):        
        dom = parseString(Grabber(url).content)
        for marker in dom.getElementsByTagName('marker'):
            values = xml_station_information_wrapper(marker)
            station = StationInformation.update_or_create(values, surrogate=False)

        session.commit()

    @staticmethod
    def flush():
        StationInformation.query.delete()




"""Unit tests for Station object"""
import unittest

from veliberator.settings import TEST_XML_URL_DATA_STATION
from veliberator.station import UnknowStation, Station
from veliberator.models import StationInformation
from veliberator import Cartography

from veliberator.station import STATUS_OPEN, STATUS_CLOSE, STATUS_BONUS, \
     STATUS_ERROR, STATUS_BIKE_ONLY, STATUS_PARKING_ONLY, STATUS_NO_SERVICE, \
     STATUS_ALMOST_EMPTY, STATUS_ALMOST_FULL

class StationTestCase(unittest.TestCase):

    def setUp(self):
        self.velib_id = 42008
        self.informations = StationInformation(id=self.velib_id,
                                               address='Test',
                                               bonus=True, opened=True)
        self.informations.save()

    def tearDown(self):
        self.informations = None
        StationInformation.query.delete()

    def test_InformationsErrorRaising(self):
        self.assertRaises(UnknowStation, Station, 0)

    def test_InformationsNoError(self):
        station = Station(self.velib_id)
        self.assertTrue(isinstance(station.informations, StationInformation))
        station = Station(str(self.velib_id))
        self.assertTrue(isinstance(station.informations, StationInformation))

    def test_Properties(self):
        station = Station(self.velib_id)
        self.assertTrue(station.is_open == station.informations.opened)
        self.assertTrue(station.is_bonus == station.informations.bonus)

    def test_IsFree(self):
        station = Station(self.velib_id)
        station.status.status['free'] = 5
        self.assertTrue(station.is_free())
        self.assertTrue(station.is_free(4))
        self.assertFalse(station.is_free(7))

    def test_IsAvailable(self):
        station = Station(self.velib_id)
        station.status.status['available'] = 5
        self.assertTrue(station.is_available())
        self.assertTrue(station.is_available(4))
        self.assertFalse(station.is_available(7))

    def test_GetStationsAround(self):
        Cartography.synchronize(TEST_XML_URL_DATA_STATION)
        
        station = Station(self.velib_id)

        stations_around = station.stations_around
        self.assertEquals(len(stations_around), 7)

        Cartography.flush()

    def test_State(self):
        status = {'total': 25, 'available': 0,
                  'free': 12, 'ticket': True}

        station = Station(self.velib_id)
        station.status.status = status

        self.assertEquals(station.state, STATUS_PARKING_ONLY)
        station.status.status['available'] = 2
        self.assertEquals(station.state, STATUS_ALMOST_EMPTY)
        station.status.status['free'] = 2
        self.assertEquals(station.state, STATUS_ALMOST_FULL)
        station.status.status['free'] = 0
        self.assertEquals(station.state, STATUS_BIKE_ONLY)
        station.status.status['available'] = 0
        self.assertEquals(station.state, STATUS_NO_SERVICE)
        station.status.status['total'] = 0
        self.assertEquals(station.state, STATUS_ERROR)
        station.status.status['total'] = 25
        station.status.status['free'] = station.status.status['available'] = 10
        self.assertEquals(station.state, STATUS_BONUS)
        station.informations.bonus = False
        self.assertEquals(station.state, STATUS_OPEN)
        station.informations.opened = False
        self.assertEquals(station.state, STATUS_CLOSE)

        
suite = unittest.TestLoader().loadTestsFromTestCase(StationTestCase)



"""Unit tests for xml functions"""
import unittest
from xml.dom.minidom import parseString

from veliberator.xml_wrappers import xml_station_status_wrapper
from veliberator.xml_wrappers import xml_station_information_wrapper


class XmlFuncTestCase(unittest.TestCase):

    def test_XmlStationStatusWrapper(self):
        data_reference = {'total': 25, 'available': 23,
                          'free': 1, 'ticket': True}

        xml = parseString('<station>' \
                          '<available>23</available> <free>1</free>' \
                          '<total>25</total> <ticket>1</ticket>' \
                          '</station>')
        self.assertEquals(xml_station_status_wrapper(xml.firstChild), data_reference)

    def test_XmlStationInformationWrapper(self):
        data_reference = {'bonus': False, 'opened': True,
                          'postal_code': u'75020', 'city': u'PARIS', 
                          'address': u'69 RUE SAINT BLAISE',
                          'lat': u'48.8568139855',
                          'lng': u'2.40903293016',
                          'id': 20017}

        xml = parseString('<marker name="20017 - RUE SAINT BLAISE" number="20017" ' \
                          'address="69 RUE SAINT BLAISE -" ' \
                          'fullAddress="69 RUE SAINT BLAISE - 75020 PARIS" ' \
                          'lat="48.8568139855" lng="2.40903293016" open="1" bonus="0"/>')
        
        self.assertEquals(xml_station_information_wrapper(xml.firstChild), data_reference)

    def test_XmlStationInformationWrapperAdvanced(self):
        data_reference = {'bonus': False, 'opened': True,
                          'postal_code': u'94200', 'city': u'IVRY', 
                          'address': u'157-165 AVENUE DE VERDUN',
                          'lat': u'48.8067594749',
                          'lng': u'2.37550404031',
                          'id': 42009}

        xml = parseString('<marker name="42009 - VERDUN (IVRY)" number="42009" ' \
                          'address="157-165 AVENUE DE VERDUN -" ' \
                          'fullAddress="157-165 AVENUE DE VERDUN - 94200 IVRY" ' \
                          'lat="48.8067594749" lng="2.37550404031" open="1" bonus="0"/>')
        
        self.assertEquals(xml_station_information_wrapper(xml.firstChild), data_reference)


suite = unittest.TestLoader().loadTestsFromTestCase(XmlFuncTestCase)



#!/usr/bin/python
import sys, os
from optparse import OptionParser

import sqlalchemy

sys.path.append(os.path.abspath('.'))

from veliberator import Station
from veliberator import VERSION
from veliberator import Cartography
from veliberator.models import db_connection
from veliberator.settings import DATABASE_URI
from veliberator.geofinder import GeoFinderError
from veliberator.geofinder import AddressGeoFinder
from veliberator.models import StationInformation

def check_database_content():
    if not StationInformation.query.count():
        print '-> Synchronisation en ligne...'
        Cartography.synchronize()
        print '-> Synchronisation complete !'

def show_status(station, distance=None):
    print "Station '%s'" % station.id
    if distance:
        print '%s (%.2fm)' % (station.informations.full_address, distance)
    else:
        print station.informations.full_address
    print '%s/%s velo(s) disponible' % (station.status.available, station.status.total)
    print '%s place(s) disponible' % station.status.free

def display_free_stations(stations, places, max_display):
    displayed = 0

    for station_information in stations:
        if displayed == max_display:
            break
        station = Station(station_information.id)
        if station.is_free(places):
            print '----------'
            show_status(station, station_information.distance)
            displayed += 1
    

if __name__ == '__main__':   
    parser = OptionParser(usage='usage: %prog [station_id] [options]')
    parser.add_option('-d', '--database', dest='database',
                      help='The SQLURI of the database', default=DATABASE_URI)
    parser.add_option('-p', '--places', dest='places', type='int',
                      help='The number of places you want', default=1)
    parser.add_option('-m', '--max_stations', dest='max_stations', type='int',
                      help='The maximun stations around proposed', default=5)
    (options, args) = parser.parse_args()

    print '-==* Veliberator Find Place v%s *==-' % VERSION

    try:
        db_connection(options.database)
    except sqlalchemy.exc.OperationalError:
        print '-> Database innacessible, switch sur la RAM.'
        print '-> Modifiez le fichier de configuration, pour enlever ce message.'
        db_connection('sqlite://')

    check_database_content()
    
    if args:
        user_input = args[0]
    else:
        user_input = raw_input('Identifiant de la station, ou adresse complete :\n')

    if user_input.isdigit():
        station = Station(user_input)
        show_status(station)
        if not station.is_free(options.places):
            print 'Calcul des places les plus proches...'
            display_free_stations(station.stations_around,
                                  options.places, options.max_stations)
    else:
        try:
            finder = AddressGeoFinder(user_input)            
        except GeoFinderError:
            sys.exit('Votre addresse n\'est pas valide ou imprecise.')
        display_free_stations(finder.get_stations_around(),
                              options.places, options.max_stations)



"""Unit tests for Status object"""
import unittest

from veliberator.grabber import Grabber
from veliberator.settings import PROXY_SERVERS
from veliberator.settings import TEST_XML_URL_DATA_STATION

class GrabberTestCase(unittest.TestCase):

    def test_Init(self):
        grabber = Grabber(TEST_XML_URL_DATA_STATION)
        self.assertEquals(grabber.url, TEST_XML_URL_DATA_STATION)
        self.assertEquals(grabber.proxies, PROXY_SERVERS)

    def test_ContentAndCache(self):
        grabber = Grabber(TEST_XML_URL_DATA_STATION)
        self.assertEquals(grabber.data, None)
        self.assertEquals(grabber.page, None)

        content = grabber.content
        page = grabber.page
        self.assertNotEquals(grabber.data, None)
        self.assertNotEquals(grabber.page, None)
        self.assertEquals(grabber.data, content)
        #Normaly does not reload the page when reaccessing to content
        self.assertEquals(grabber.content, content)
        self.assertEquals(grabber.page, page)

    def test_ErrorGrabbing(self):
        grabber = Grabber('http://example.com/badurl')
        self.assertEquals(grabber.content, '')

suite = unittest.TestLoader().loadTestsFromTestCase(GrabberTestCase)



"""Grabber for collecting data"""
import urllib2
from random import sample

from veliberator.settings import PROXY_SERVERS

class Grabber(object):
    """Url encapsultation for making request throught HTTP"""
    page = None
    data = None

    def __init__(self, url, proxies=PROXY_SERVERS):
        """Init the grabber"""
        self.url = url
        self.proxies = proxies
        self.opener = self.build_opener()

    def build_opener(self):
        """Build the url opener"""
        handlers = []
        
        if self.proxies:
            server = sample(self.proxies, 1)[0]
            handlers.append(urllib2.ProxyHandler({"http" : server}))

        return urllib2.build_opener(*handlers)

    @property
    def content(self):
        """Return the data grabbed"""
        if self.data:
            return self.data
        try:
            self.page = self.opener.open(self.url)
            self.data = ''.join(self.page.readlines())
            self.page.close()        
            return self.data
        except IOError, e:
            return ''



"""Installing veliberator"""
from setuptools import setup, find_packages
import sys, os

from veliberator import VERSION

setup(
    name='veliberator',
    version=VERSION,
    zip_safe=True,

    scripts=['./bin/find_place.py',
             './bin/synchronize.py'],
    data_files=[('/etc', ['etc/veliberator.cfg']),],

    packages=find_packages(exclude=['tests',]),
    install_requires = ['SQLAlchemy>=0.5',
                        'Elixir>=0.7.0',
                        'simplejson>=2.0.9'],
    include_package_data=True,
        
    test_suite = 'tests.global_test_suite',

    author='Fantomas42',
    author_email='fantomas42@gmail.com',
    url='http://veliberator.com',
 
    license='GPL',
    platforms = 'any',
    description='Python API for Velib.',
    long_description=open(os.path.join('docs', 'README.txt')).read(),
    keywords='velib, api, service',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )



"""Xml functions for helping in convertion of data"""

def xml_station_status_wrapper(xmlnode):
    """Convert Station status xml
    to a usable dict"""
    
    return {'total': int(xmlnode.getElementsByTagName('total')[0].childNodes[0].data),
            'available': int(xmlnode.getElementsByTagName('available')[0].childNodes[0].data),
            'free': int(xmlnode.getElementsByTagName('free')[0].childNodes[0].data),
            'ticket': int(xmlnode.getElementsByTagName('ticket')[0].childNodes[0].data) == 1}
 

def xml_station_information_wrapper(xmlnode):
    """Convert Station information xml
    to a usable dict"""
    city = ''
    postal_code = ''
    address = xmlnode.getAttribute('address')[:-1].strip()
    address_parts = xmlnode.getAttribute('fullAddress').split()

    for p in address_parts:
        if len(p) == 5 and p.isdigit():
            postal_code = p
            city = ' '.join(address_parts[address_parts.index(p) + 1:])
            break

    return {'id': int(xmlnode.getAttribute('number')),
            'address': address,
            'postal_code': postal_code,
            'city': city,
            'lat': xmlnode.getAttribute('lat'),
            'lng': xmlnode.getAttribute('lng'),
            'opened': xmlnode.getAttribute('open') == '1',
            'bonus': xmlnode.getAttribute('bonus') == '1',}



"""Geo localizator objects"""
from math import sqrt, radians, sin, cos, atan2
from urllib import quote_plus
from urllib import urlopen

import simplejson as json

from veliberator.models import StationInformation

global_geofinder_cache = {}

def cache_wrapper(method):
    def cache(instance):
        """Use a global cache for the results
        of stations around"""
        global global_geofinder_cache
        
        key_cache = (instance.lat, instance.lng)
        if not key_cache in global_geofinder_cache.keys():
            global_geofinder_cache[key_cache] = method(instance)
       
        return global_geofinder_cache[key_cache]
    return cache

def pythagor_distance(start, end):
    """Compute the distance between 2 points
    with Pythagor theorem"""
    l1 = float(start[0]) - float(end[0])
    l2 = float(start[1]) - float(end[1])
    return sqrt(pow(l1, 2) + pow(l2, 2))

def haversine_distance(start, end):
    """Compute the distance between 2 points in meters
    with the haversine formula"""
    radius = float(6378137) # Earth radius in meters

    start_long = radians(float(start[0]))
    start_latt = radians(float(start[1]))
    end_long = radians(float(end[0]))
    end_latt = radians(float(end[1]))
    d_latt = end_latt - start_latt
    d_long = end_long - start_long
    a = sin(d_latt/2)**2 + cos(start_latt) * cos(end_latt) * sin(d_long/2)**2
    c = 2 * atan2(sqrt(a), sqrt(1-a))

    return radius * c

class GeoFinderError(Exception):
    pass

class BaseGeoFinder(object):
    """Base GeoFinder object,
    for finding stations around"""
    lat = None
    lng = None
    
    PRECISION = '%.2f'
    SQUARE_SIZE = 0.01

    def __init__(self, lat, lng):
        self.lat = lat
        self.lng = lng

    def compute_square_area(self):
        """Round the GPS coordonates to a wide area"""
        lat_orig = float(self.lat)        
        lat_pos = self.PRECISION % (lat_orig + self.SQUARE_SIZE)
        lat_neg = self.PRECISION % (lat_orig - self.SQUARE_SIZE)
        lat_orig = self.PRECISION % lat_orig

        lng_orig = float(self.lng)
        lng_pos = self.PRECISION % (lng_orig + self.SQUARE_SIZE)
        lng_neg = self.PRECISION % (lng_orig - self.SQUARE_SIZE)
        lng_orig = self.PRECISION % lng_orig

        return (lat_neg, lat_orig, lat_pos), (lng_neg, lng_orig, lng_pos)

    def compute_station_distances(self, stations,
                                  distance_function=haversine_distance):
        """Return the stations ordered by their distance to the
        origin, with a distance property added"""
        station_distances = {}

        for station in stations:
            if station.lat == self.lat and station.lng == self.lng:
                continue
            distance = distance_function((self.lat, self.lng),
                                         (station.lat, station.lng))
            station.distance = distance
            station_distances[station] = distance

        stations_sorted_by_distance = sorted(station_distances.iteritems(),
                                             key=lambda (k,v): (v,k))
        return [station for station, distance in stations_sorted_by_distance]

    def get_stations_in_area(self, lats, lngs):
        """Make a query to find the station with
        lat and lng matching with the area"""
        return StationInformation.query.filter(
            (StationInformation.lat >= lats[0]) &
            (StationInformation.lat <= lats[2]) &
            (StationInformation.lng >= lngs[0]) &
            (StationInformation.lng <= lngs[2])).all()

    @cache_wrapper
    def get_stations_around(self):
        """Find the stations around the lat and lng,
        sorted by proximity"""
        lats, lngs = self.compute_square_area()
        stations = self.get_stations_in_area(lats, lngs)
        stations_distanced = self.compute_station_distances(stations)

        return stations_distanced

class StationGeoFinder(BaseGeoFinder):

    def __init__(self, station):
        """Init the object with a Station object"""
        self.lat = station.informations.lat
        self.lng = station.informations.lng


class AddressGeoFinder(BaseGeoFinder):

    def __init__(self, address):
        """Init the object with an address
        who will be geocomputed"""
        self.address = address
        informations = self.geocompute(address)

        self.lat = informations['Point']['coordinates'][1]
        self.lng = informations['Point']['coordinates'][0]
        self.precision = informations['AddressDetails']['Accuracy']
        self.clean_address = informations['address']   

        if self.precision <= 6:
            raise GeoFinderError('Your address is not available.')

    def geocompute(self, address):
        """Geocompute an address with GMap"""
        address = quote_plus(address)
        request = "http://maps.google.com/maps/geo?sensor=false&q=%s&output=%s&oe=utf8&gl=fr" % (
            address, 'json')
        data = json.loads(urlopen(request).read())
        
        if data['Status']['code'] != 200:
            raise GeoFinderError('Service unavailable')
        return data['Placemark'][0]





"""Models for database"""
from datetime import datetime

from elixir import *

from veliberator.settings import DATABASE_URI, DATABASE_ECHO

class StationInformation(Entity):
    """Station information entity model"""
    id = Field(Integer, primary_key=True)

    address = Field(String(100))
    postal_code = Field(String(100))
    city = Field(String(100))
    
    lat = Field(String(30))
    lng = Field(String(30))

    opened = Field(Boolean)
    bonus = Field(Boolean)

    creation_date = Field(DateTime, default=datetime.now)
    update = Field(DateTime, default=datetime.now)

    using_options(order_by='id',
                  tablename='veliberator_station_information')

    def __repr__(self):
        return '<StationInformation "%s" (%s)>' % (self.id, self.address)

    @property
    def full_address(self):
        return '%s, %s %s' % (self.address,
                              self.postal_code,
                              self.city)


def db_connection(sqluri=DATABASE_URI, echo=DATABASE_ECHO):
    metadata.bind = sqluri
    metadata.bind.echo = echo
    setup_all()
    create_all()



"""Unit tests for Status object"""
import unittest
from datetime import datetime, timedelta

from veliberator.status import StationStatus
from veliberator.status import global_stationstatus_cache
from veliberator.settings import STATION_STATUS_RECENT

class StationStatusTestCase(unittest.TestCase):

    def setUp(self):
        self.velib_id = 42008

    def tearDown(self):
        global global_stationstatus_cache
        global_stationstatus_cache = {}

    def test_Init(self):
        status = StationStatus(self.velib_id)
        self.assertTrue(isinstance(status.ticket, bool))
        self.assertTrue(isinstance(status.total, int))
        self.assertTrue(isinstance(status.available, int))
        self.assertTrue(isinstance(status.free, int))

    def test_Cache(self):
        global global_stationstatus_cache
        
        status = StationStatus(self.velib_id)
        data_compare = global_stationstatus_cache[self.velib_id].copy()
        status.get_status()
        self.assertEquals(status.status, data_compare)
        global_stationstatus_cache[self.velib_id][\
            'datetime'] = datetime.now() - timedelta(minutes=STATION_STATUS_RECENT + 5)
        status.get_status()
        self.assertNotEquals(status.status, data_compare)

    def test_GetStatus(self):
        status = StationStatus(self.velib_id)
        self.assertTrue(isinstance(status.get_status(), dict))

    def test_GetStatusError(self):
        status = StationStatus(self.velib_id - 1, 'http://example.com/badurl')
        self.assertTrue(isinstance(status.get_status(), dict))
        self.assertEquals(status.free, 0)

suite = unittest.TestLoader().loadTestsFromTestCase(StationStatusTestCase)

    



"""Unit tests for models object"""
import unittest

from veliberator.models import StationInformation, session


class StationInformationTestCase(unittest.TestCase):

    def tearDown(self):
        StationInformation.query.delete()

    def test_UpdateOrCreate(self):
        self.assertEquals(StationInformation.query.count(), 0)
        
        values = {'id': 42,
                  'address': '1 rue Kennedy',
                  'postal_code': '94200',
                  'city': 'Paris',
                  'lat': '42.000000000',
                  'lng': '42.000000000',
                  'opened': True,
                  'bonus': False}
        station = StationInformation.update_or_create(values, surrogate=False)
        self.assertEquals(StationInformation.query.count(), 1)
        self.assertEquals(station.postal_code, '94200')

        values['postal_code'] = '75000'
        station = StationInformation.update_or_create(values, surrogate=False)
        self.assertEquals(StationInformation.query.count(), 1)
        self.assertEquals(station.postal_code, '75000')
        
        values['id'] = 43
        station = StationInformation.update_or_create(values, surrogate=False)
        self.assertEquals(StationInformation.query.count(), 2)
        self.assertEquals(station.postal_code, '75000')

    def test_FullAddress(self):
        values = {'id': 42,
                  'address': '1 rue Kennedy',
                  'postal_code': '94200',
                  'city': 'Paris',
                  'lat': '42.000000000',
                  'lng': '42.000000000',
                  'opened': True,
                  'bonus': False}
        station = StationInformation.update_or_create(values, surrogate=False)
        self.assertEquals(station.full_address, '1 rue Kennedy, 94200 Paris')
        
suite = unittest.TestLoader().loadTestsFromTestCase(StationInformationTestCase)
